import pytest
from gaarf.base_query import BaseQuery


@pytest.fixture
def fake_query():
    class FakeQuery(BaseQuery):
        def __init__(self):
            self.query_text = "SELECT 1"
    fake_query = FakeQuery()
    return fake_query


@pytest.fixture
def wrong_query():
    class WrongQuery(BaseQuery):
        def __init__(self):
            self.query = "SELECT 1"
    wrong_query = WrongQuery()
    return wrong_query


def test_base_query_init_raises_error():
    with pytest.raises(NotImplementedError):
        base_query = BaseQuery()


def test_implemented_str(fake_query):
    assert str(fake_query) == "SELECT 1"


def test_not_implemented_str(wrong_query):
    with pytest.raises(NotImplementedError):
        print(wrong_query)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional
from pathlib import Path
from google.ads.googleads.client import GoogleAdsClient  # type: ignore


class BaseClient:

    def get_response(self, entity_id: str, query_text: str):
        pass


class GoogleAdsApiClient(BaseClient):
    default_google_ads_yaml = str(Path.home() / "google-ads.yaml")

    def __init__(self,
                 path_to_config: str = default_google_ads_yaml,
                 config_dict: Dict[str, Any] = None,
                 yaml_str: str = None,
                 version: str = "v9"):
        self.client = self._init_client(path=path_to_config,
                                        config_dict=config_dict,
                                        yaml_str=yaml_str,
                                        version=version)
        self.ads_service = self.client.get_service("GoogleAdsService")

    def get_response(self, entity_id, query_text):
        response = self.ads_service.search_stream(customer_id=entity_id,
                                                  query=query_text)
        return response

    def _init_client(self, path, config_dict, yaml_str,
                     version) -> Optional[GoogleAdsClient]:
        if path:
            return GoogleAdsClient.load_from_storage(path, version)
        if config_dict:
            return GoogleAdsClient.load_from_dict(config_dict, version)
        if yaml_str:
            return GoogleAdsClient.load_from_string(yaml_str, version)
        try:
            return GoogleAdsClient.load_from_env(version)
        except Exception as e:
            raise ValueError("Cannot instantiate GoogleAdsClient")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

from concurrent import futures
import sys
import argparse
from pathlib import Path
import logging
import traceback

from google.ads.googleads.errors import GoogleAdsException  # type: ignore

from gaarf import api_clients, utils, query_executor
from gaarf.io import writer, reader  # type: ignore
from .utils import GaarfConfigBuilder, ConfigSaver, initialize_runtime_parameters


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--account", dest="customer_id")
    parser.add_argument("--output", dest="save", default="csv")
    parser.add_argument("--input", dest="input", default="file")
    parser.add_argument("--ads-config",
                        dest="config",
                        default=str(Path.home() / "google-ads.yaml"))
    parser.add_argument("--api-version", dest="api_version", default=10)
    parser.add_argument("--log",
                        "--loglevel",
                        dest="loglevel",
                        default="warning")
    parser.add_argument("--customer-ids-query",
                        dest="customer_ids_query",
                        default=None)
    parser.add_argument("--customer-ids-query-file",
                        dest="customer_ids_query_file",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.set_defaults(save_config=False)
    args = parser.parse_known_args()
    main_args = args[0]

    config = GaarfConfigBuilder(args).build()

    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)

    config = initialize_runtime_parameters(config)

    logging.basicConfig(
        format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
        stream=sys.stdout,
        level=main_args.loglevel.upper(),
        datefmt="%Y-%m-%d %H:%M:%S")
    logging.getLogger("google.ads.googleads.client").setLevel(logging.WARNING)

    ads_client = api_clients.GoogleAdsApiClient(
        path_to_config=main_args.config, version=f"v{config.api_version}")
    writer_client = writer.WriterFactory().create_writer(
        config.output, **config.writer_params)
    reader_factory = reader.ReaderFactory()
    reader_client = reader_factory.create_reader(main_args.input)

    if main_args.customer_ids_query:
        console_reader = reader_factory.create_reader("console")
        customer_ids_query = console_reader.read(main_args.customer_ids_query)
    elif main_args.customer_ids_query_file:
        file_reader = reader_factory.create_reader("file")
        customer_ids_query = file_reader.read(
            main_args.customer_ids_query_file)
    else:
        customer_ids_query = None

    customer_ids = utils.get_customer_ids(ads_client, config.account,
                                          customer_ids_query)
    ads_query_executor = query_executor.AdsQueryExecutor(ads_client)

    logging.info("Total number of customer_ids is %d, accounts=[%s]",
                 len(customer_ids), ",".join(map(str, customer_ids)))

    with futures.ThreadPoolExecutor() as executor:
        future_to_query = {
            executor.submit(ads_query_executor.execute, query, customer_ids,
                            reader_client, writer_client, config.params): query
            for query in main_args.query
        }
        for future in futures.as_completed(future_to_query):
            query = future_to_query[future]
            try:
                future.result()
                print(f"{query} executed successfully")
            except writer.ZeroRowException:
                print(f"""
                    {query} generated ZeroRowException,
                    please check WHERE statements.
                    """)
            except GoogleAdsException as ex:
                print(f'"{query}" failed with status '
                      f'"{ex.error.code().name}" and includes'
                      'the following errors:')
                for error in ex.failure.errors:
                    print(f'\tError with message "{error.message}".')
                    if error.location:
                        for field in error.location.field_path_elements:
                            print(f"\t\tOn field: {field.field_name}")
            except Exception as e:
                traceback.print_tb(e.__traceback__)
                print(f"{query} generated an exception: {str(e)}")


if __name__ == "__main__":
    main()



import pytest
from google.cloud.bigquery import SchemaField  # type: ignore
import gaarf.io.writer as writer  # type: ignore
from gaarf.report import GaarfReport


@pytest.fixture
def csv_writer():
    return writer.CsvWriter("/tmp")


@pytest.fixture
def bq_writer():
    return writer.BigQueryWriter("fake_project", "fake_dataset")


@pytest.fixture
def single_column_data():
    results = [1, 2, 3]
    columns = ["column_1"]
    return GaarfReport(results, columns)


@pytest.fixture
def sample_data():
    results = [(1, "two", [3, 4])]
    columns = ["column_1", "column_2", "column_3"]
    return GaarfReport(results, columns)


def test_csv_writer_single_column(csv_writer, single_column_data):
    tmp_file = "/tmp/test.csv"
    expected = ["column_1", "1", "2", "3"]
    csv_writer.write(single_column_data, "test.csv")
    with open(tmp_file, "r") as f:
        file = f.readlines()
    assert [row.strip() for row in file] == expected


def test_csv_writer_multi_column(csv_writer, sample_data):
    tmp_file = "/tmp/test.csv"
    expected = ["column_1,column_2,column_3", '1,two,3|4']
    csv_writer.write(sample_data, "test.csv")
    with open(tmp_file, "r") as f:
        file = f.readlines()
    assert [row.strip() for row in file] == expected


def test_bq_get_results_types(bq_writer, sample_data):
    results, columns = sample_data.results, sample_data.column_names
    result_types = bq_writer._get_result_types(results, columns)
    assert result_types == {
        'column_1': {
            'element_type': int,
            'repeated': False
        },
        'column_2': {
            'element_type': str,
            'repeated': False
        },
        'column_3': {
            'element_type': int,
            'repeated': True
        }
    }


def test_bq_get_correct_schema(bq_writer, sample_data):
    results, columns = sample_data.results, sample_data.column_names
    schema = bq_writer._define_schema(results, columns)
    assert schema == [
        SchemaField('column_1', 'INT64', 'NULLABLE', None, (), None),
        SchemaField('column_2', 'STRING', 'NULLABLE', None, (), None),
        SchemaField('column_3', 'INT64', 'REPEATED', None, (), None)
    ]


def test_format_extension():
    default_output = writer.DestinationFormatter.format_extension(
        "test_query.sql")
    default_output_custom_extension = writer.DestinationFormatter.format_extension(
        "test_query.txt", ".txt")
    csv_output = writer.DestinationFormatter.format_extension(
        "test_query.sql", new_extension=".csv")
    assert default_output == "test_query"
    assert default_output_custom_extension == "test_query"
    assert csv_output == "test_query.csv"


@pytest.fixture
def writer_factory():
    return writer.WriterFactory()


def test_writer_factory_load(writer_factory):
    assert writer_factory.write_options == {
        "bq": writer.BigQueryWriter,
        "csv": writer.CsvWriter,
        "console": writer.StdoutWriter,
        "sqldb": writer.SqlAlchemyWriter
    }


def test_writer_factory_inits(writer_factory):
    bq_writer = writer_factory.create_writer("bq",
                                             project="fake_project",
                                             dataset="fake_dataset")
    csv_writer = writer_factory.create_writer(
        "csv", destination_folder="/fake_folder")
    sqlalchemy_writer = writer_factory.create_writer(
        "sqldb",
        connection_string="protocol://user:password@host:port/db")
    assert bq_writer.dataset_id == "fake_project.fake_dataset"
    assert csv_writer.destination_folder == "/fake_folder"
    assert sqlalchemy_writer.connection_string == "protocol://user:password@host:port/db"


def test_null_writer_raises_unknown_writer_error(writer_factory):
    with pytest.raises(ValueError):
        writer_factory.create_writer("non-existing-option")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Tuple, Sequence, Union
from operator import attrgetter
import re
import proto  # type: ignore
from proto.marshal.collections.repeated import (  # type: ignore
    Repeated, RepeatedComposite)  # type: ignore


class BaseParser:
    def __init__(self, successor):
        self._successor = successor

    def parse(self, request):
        if self._successor:
            return self._successor.parse(request)
        return None


class RepeatedParser(BaseParser):
    def parse(self, request):
        if isinstance(request, Repeated) and "customer" in str(request):
            elements = []
            for request_element in request:
                element = ResourceFormatter.get_resource_id(request_element)
                elements.append(ResourceFormatter.clean_resource_id(element))
            return elements
        return super().parse(request)


class RepeatedCompositeParser(BaseParser):
    def parse(self, request):
        if isinstance(request, RepeatedComposite):
            elements = []
            for request_element in request:
                resource = ResourceFormatter.get_resource(request_element)
                element = ResourceFormatter.get_resource_id(resource)
                elements.append(ResourceFormatter.clean_resource_id(element))
            return elements
        return super().parse(request)


class AttributeParser(BaseParser):
    def parse(self, request):
        if hasattr(request, "name"):
            return request.name
        if hasattr(request, "text"):
            return request.text
        if hasattr(request, "value"):
            return request.value
        return super().parse(request)


class EmptyAttributeParser(BaseParser):
    def parse(self, request):
        if issubclass(type(request), proto.Message):
            return "Not set"
        return super().parse(request)


class GoogleAdsRowParser:
    def __init__(self, nested_fields=None):
        self.nested_fields = nested_fields
        self.parser = self._init_parsers()

    def _init_parsers(self):
        parser_chain = BaseParser(None)
        for parser in [
                EmptyAttributeParser, AttributeParser, RepeatedCompositeParser,
                RepeatedParser
        ]:
            new_parser = parser(parser_chain)
            parser_chain = new_parser
        return parser_chain

    def parse(self, request):
        return self.parser.parse(request)

    def parse_ads_row(self, row,
                      query_specification) -> Sequence[Any]:
        final_rows = []
        extracted_rows = self._get_attributes_from_row(row,
                                                       query_specification.fields)
        customizers = query_specification.customizers
        for i, r in enumerate(extracted_rows):
            if customizers:
                if customizers.get(i):
                    caller = customizers.get(i)
                    if caller.get("type") == "nested_field":
                        try:
                            r = attrgetter(caller.get("value"))(r)
                        except AttributeError:
                            raise ValueError(f"{caller} is incorrect")
                    elif caller.get("type") == "resource_index":
                        r = re.split("~", r)[caller.get("value")]
                        r = re.split("/", r)[-1]
            parsed_element = self.parser.parse(r) or r
            final_rows.append(parsed_element)
        return final_rows if len(final_rows) > 1 else final_rows[0]

    def _get_attributes_from_row(
            self, row, fields) -> Tuple[Any, ...]:
        getter = attrgetter(*fields)
        rows = getter(row)
        return rows if isinstance(rows, tuple) else (rows, )


class ResourceFormatter:
    @staticmethod
    def get_resource(element: str) -> str:
        return re.split(": ", str(element).strip())[1]

    @staticmethod
    def get_resource_id(element: str) -> str:
        return re.split("/", str(element))[-1]

    @staticmethod
    def clean_resource_id(element: str) -> Union[int, str]:
        element = re.sub('"', '', str(element))
        try:
            return int(element)
        except ValueError:
            return element



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union
from google.cloud import bigquery  # type: ignore
from google.cloud.exceptions import NotFound  # type: ignore
from jinja2 import Template


class BigQueryExecutor:
    def __init__(self, project_id: str):
        self.project_id = project_id
        self.client = bigquery.Client(project_id)

    def execute(self, script_name: str, query_text: str,
                params: Optional[Dict[str, Any]]) -> None:
        print(params)
        if params:
            if (templates := params.get("template")):
                query_text = self._expand_jinja(query_text,
                                                **templates)
            if (macros := params.get("macro")):
                query_text = query_text.format(**macros)
        job = self.client.query(query_text)
        try:
            job.result()
            print(f"{script_name} launched successfully")
        except Exception as e:
            print(f"Error launching {script_name} query!" f"{str(e)}")

    def create_datasets(self, datasets: Union[str, List[str]]) -> None:
        if isinstance(datasets, str):
            datasets = [datasets]
        for dataset in datasets:
            dataset_id = f"{self.project_id}.{dataset}"
            try:
                bq_dataset = self.client.get_dataset(dataset_id)
            except NotFound:
                bq_dataset = bigquery.Dataset(dataset_id)
                bq_dataset = self.client.create_dataset(bq_dataset, timeout=30)

    def _expand_jinja(self, query_text, **template_params):
        for key, value in template_params.items():
            if len(splitted_param := value.split(",")) > 1:
                template_params[key] = splitted_param
            else:
                template_params[key] = [value]
        template = Template(query_text)
        return template.render(template_params)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import csv

from typing import Any, Dict, Sequence, Tuple
import abc
import os
import proto  # type: ignore
from google.cloud import bigquery  # type: ignore
from google.cloud.exceptions import NotFound  # type: ignore
from pathlib import Path
import csv
import pandas as pd  # type: ignore
from sqlalchemy import create_engine
from tabulate import tabulate
from ..report import GaarfReport
from .formatter import ArrayFormatter, ResultsFormatter  # type: ignore


class AbsWriter(abc.ABC):

    @abc.abstractmethod
    def write(self, results: GaarfReport, destination: str) -> str:
        pass

    def get_columns_results(
            self, results: GaarfReport) -> Tuple[Sequence[str], Sequence[Any]]:
        column_names = results.column_names
        results = ArrayFormatter.format(
            ResultsFormatter.format(results.to_list()), "|")
        return column_names, results


class StdoutWriter(AbsWriter):

    def __init__(self, page_size=10, **kwargs):
        self.page_size = page_size

    def write(self, results, destination):
        column_names, results = self.get_columns_results(results)
        results_generator = self._paginate_rows(results, self.page_size)
        message = f"showing results for query {destination}"
        print("=" * len(message))
        print(message)
        print("=" * len(message))
        print(tabulate(next(results_generator), headers=column_names))

    def _define_header(self, results, header):
        pass

    def _paginate_rows(self, results, page_size):
        for i in range(0, len(results), page_size):
            print(f"showing rows {i+1}-{i+page_size} out of total "
                  f"{len(results)} rows")
            yield results[i:(i + page_size)]


class CsvWriter(AbsWriter):

    def __init__(self,
                 destination_folder=os.getcwd(),
                 delimiter=",",
                 quotechar='"',
                 quoting=csv.QUOTE_MINIMAL,
                 **kwargs):
        self.destination_folder = destination_folder
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.quoting = quoting

    def __str__(self):
        return f"[CSV] - data are saved to {self.destination_folder} destination_folder."

    def write(self, results, destination) -> str:
        column_names, results = self.get_columns_results(results)
        destination = DestinationFormatter.format_extension(
            destination, new_extension=".csv")
        if not os.path.isdir(self.destination_folder):
            os.makedirs(self.destination_folder)
        with open(os.path.join(self.destination_folder, destination),
                  "w") as file:
            writer = csv.writer(file,
                                delimiter=self.delimiter,
                                quotechar=self.quotechar,
                                quoting=self.quoting)
            writer.writerow(column_names)
            writer.writerows(results)
        return f"[CSV] - at {destination}"


class BigQueryWriter(AbsWriter):

    def __init__(self,
                 project: str,
                 dataset: str,
                 location: str = "US",
                 **kwargs):
        self.client = bigquery.Client()
        self.dataset_id = f"{project}.{dataset}"
        self.location = location

    def __str__(self):
        return f"[BigQuery] - {self.dataset_id} at {self.location} location."

    def write(self, results, destination) -> str:
        column_names, results = self.get_columns_results(results)
        schema = self._define_schema(results, column_names)
        self._create_or_get_dataset()
        destination = DestinationFormatter.format_extension(destination)
        table = self._create_or_get_table(f"{self.dataset_id}.{destination}",
                                          schema)
        job_config = bigquery.LoadJobConfig(
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
            schema=schema)

        df = pd.DataFrame(results, columns=column_names)
        try:
            self.client.load_table_from_dataframe(dataframe=df,
                                                  destination=table,
                                                  job_config=job_config)
        except Exception as e:
            raise ValueError(f"Unable to save data to BigQuery! {str(e)}")
        return f"[BigQuery] - at {self.dataset_id}.{destination}"

    def _define_schema(self, results, header):
        result_types = self._get_result_types(results, header)
        return self._get_bq_schema(result_types)

    def _get_result_types(
            self, elements: Sequence[Any],
            column_names: Sequence[str]) -> Dict[str, Dict[str, Any]]:
        result_types = {}
        for i, element in enumerate(elements[0]):
            element_type = type(element)
            if element_type in [
                    list, proto.marshal.collections.repeated.RepeatedComposite,
                    proto.marshal.collections.repeated.Repeated
            ]:
                repeated = True
                if len(element) == 0:
                    element_type = str
                else:
                    element_type = type(element[0])
            else:
                element_type = type(element)
                repeated = False
            result_types[column_names[i]] = {
                "element_type": element_type,
                "repeated": repeated
            }
        return result_types

    def _get_bq_schema(self, types):
        TYPE_MAPPING = {
            list: "REPEATED",
            str: "STRING",
            int: "INT64",
            float: "FLOAT64",
            bool: "BOOL",
            proto.marshal.collections.repeated.RepeatedComposite: "REPEATED",
            proto.marshal.collections.repeated.Repeated: "REPEATED"
        }

        schema = []
        for key, value in types.items():
            element_type = TYPE_MAPPING.get(value.get("element_type"))
            schema.append(
                bigquery.SchemaField(
                    name=key,
                    field_type="STRING" if not element_type else element_type,
                    mode="REPEATED" if value.get("repeated") else "NULLABLE"))
        return schema

    def _create_or_get_dataset(self):
        try:
            bq_dataset = self.client.get_dataset(self.dataset_id)
        except NotFound:
            bq_dataset = bigquery.Dataset(self.dataset_id)
            bq_dataset.location = self.location
            bq_dataset = self.client.create_dataset(bq_dataset, timeout=30)
        return bq_dataset

    def _create_or_get_table(self, table_name, schema):
        try:
            table = self.client.get_table(table_name)
        except NotFound:
            table_ref = bigquery.Table(table_name, schema=schema)
            table = self.client.create_table(table_ref)
            table = self.client.get_table(table_name)
        return table


class SqlAlchemyWriter(AbsWriter):

    def __init__(self,
                 connection_string: str,
                 if_exists: str = "replace",
                 **kwargs):
        self.connection_string = connection_string
        self.if_exists = if_exists

    def write(self, results, destination):
        column_names, results = self.get_columns_results(results)
        destination = DestinationFormatter.format_extension(destination)
        pd.DataFrame(data=results,
                     columns=column_names).to_sql(name=destination,
                                                  con=self._create_engine(),
                                                  if_exists=self.if_exists)

    def _create_engine(self):
        return create_engine(self.connection_string)


class NullWriter(AbsWriter):

    def __init__(self, writer_option, **kwargs):
        raise ValueError(f"{writer_option} is unknown writer type!")

    def write(self):
        print("Unknown writer type!")


class WriterFactory:
    write_options: Dict[str, AbsWriter] = {}

    def __init__(self):
        self.load_writer_options()

    def load_writer_options(self):
        self.write_options["bq"] = BigQueryWriter
        self.write_options["csv"] = CsvWriter
        self.write_options["console"] = StdoutWriter
        self.write_options["sqldb"] = SqlAlchemyWriter

    def create_writer(self, writer_option, **kwargs):
        if writer_option in self.write_options:
            return self.write_options[writer_option](**kwargs)
        else:
            return NullWriter(writer_option)


class DestinationFormatter:

    @staticmethod
    def format_extension(path_object: str,
                         current_extension: str = ".sql",
                         new_extension: str = "") -> str:
        return Path(path_object).name.replace(current_extension, new_extension)


class ZeroRowException(Exception):
    pass



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Sequence, Union
from google.ads.googleads.errors import GoogleAdsException  # type: ignore

from . import parsers
from . import api_clients
from .query_editor import QuerySpecification, QueryElements
from .report import GaarfReport
from .io import writer, reader  # type: ignore


class AdsReportFetcher:
    def __init__(self, api_client: api_clients.BaseClient,
                 customer_ids: Union[List[str], str]):
        self.api_client = api_client
        self.customer_ids = [
            customer_ids
        ] if not isinstance(customer_ids, list) else customer_ids

    def fetch(
        self,
        query_specification: Union[str, QueryElements],
    ) -> GaarfReport:
        total_results = []
        if not isinstance(query_specification, QueryElements):
            query_specification = QuerySpecification(
                str(query_specification)).generate()

        for customer_id in self.customer_ids:
            try:
                results = self._parse_ads_response(query_specification,
                                                   customer_id)
                total_results.extend(results)
                if query_specification.is_constant_resource:
                    print("Running only once")
                    break
            except GoogleAdsException as e:
                print("Cannot execute query for "
                      f"{query_specification.query_title} "
                      f"for customer_id {customer_id}")
                print(e)
        return GaarfReport(results=total_results,
                           column_names=query_specification.column_names)

    def _parse_ads_response(self, query_specification: QueryElements,
                            customer_id: str) -> Sequence[Any]:
        parser = parsers.GoogleAdsRowParser()
        total_results = []
        response = self.api_client.get_response(
            entity_id=str(customer_id),
            query_text=query_specification.query_text)
        for batch in response:
            results = [
                parser.parse_ads_row(row, query_specification)
                for row in batch.results
            ]
            total_results.extend(results)
        return total_results


class AdsQueryExecutor:
    def __init__(self, api_client: api_clients.BaseClient):
        """
        api_client: Client used to perform authentication to Ads API.
        """
        self.api_client = api_client

    def execute(self,
                query: str,
                customer_ids: Union[List[str], str],
                reader_client: reader.AbsReader,
                writer_client: writer.AbsWriter,
                args: Dict[Any, Any] = None) -> None:
        """Reads query, extract results and stores them in a specified location.

        Attributes:
            query: Path to a file that contains query text.
            customer_ids: All accounts for which query will be executed.
            reader_client: Client responsible for reading data from local storage
            writer_client: Client responsible for writing data to local/remote
                location.
            args: Arguments that need to be passed to the query
        """

        query_text = reader_client.read(query)
        query_specification = QuerySpecification(query_text, query,
                                                 args).generate()
        report_fetcher = AdsReportFetcher(self.api_client, customer_ids)
        results = report_fetcher.fetch(query_specification)
        if len(results) > 0:
            writer_client.write(results, query_specification.query_title)
        else:
            raise writer.ZeroRowException



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Sequence, Union
import dataclasses
import os
import datetime
from dateutil.relativedelta import relativedelta
import yaml


@dataclasses.dataclass
class GaarfConfig:
    output: str
    api_version: str
    account: str
    params: Dict[str, Any]
    writer_params: Dict[str, Any]
    customer_ids_query: str
    customer_ids_query_file: str


@dataclasses.dataclass
class GaarfBqConfig:
    project: str
    target: Sequence[str]
    params: Dict[str, Any]


class BaseConfigBuilder:

    def __init__(self, args):
        self.args = args
        if (gaarf_config_path := self.args[0].gaarf_config):
            self.type = "file"
            self.gaarf_config_path = gaarf_config_path
        else:
            self.type = "cli"
            self.gaarf_config_path = None

    def build(self) -> Union[GaarfConfig, GaarfBqConfig]:
        if self.type == "file":
            return self._load_gaarf_config()
        return self._build_gaarf_config()

    def _load_gaarf_config(self) -> Union[GaarfConfig, GaarfBqConfig]:
        raise NotImplementedError

    def _build_gaarf_config(self) -> Union[GaarfConfig, GaarfBqConfig]:
        raise NotImplementedError


class GaarfConfigBuilder(BaseConfigBuilder):

    def __init__(self, args):
        super().__init__(args)

    def _load_gaarf_config(self) -> GaarfConfig:
        with open(self.gaarf_config_path, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get("gaarf")
        if not gaarf_section:
            raise ValueError("Invalid config, must have `gaarf` section!")
        if not (output := gaarf_section.get("output")):
            raise ValueError("Config does not contains `output` section!")
        return GaarfConfig(
            output=gaarf_section.get("output"),
            api_version=gaarf_section.get("api_version"),
            account=gaarf_section.get("account"),
            params=gaarf_section.get("params"),
            writer_params=gaarf_section.get(output),
            customer_ids_query=gaarf_section.get("customer_ids_query"),
            customer_ids_query_file=gaarf_section.get(
                "customer_ids_query_file"))

    def _build_gaarf_config(self) -> GaarfConfig:
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(["macro", main_args.save]).parse(query_args)
        return GaarfConfig(
            output=main_args.save,
            api_version=main_args.api_version,
            account=main_args.customer_id,
            params=params,
            writer_params=params.get(main_args.save),
            customer_ids_query=main_args.customer_ids_query,
            customer_ids_query_file=main_args.customer_ids_query_file)


class GaarfBqConfigBuilder(BaseConfigBuilder):

    def __init__(self, args):
        super().__init__(args)

    def _load_gaarf_config(self) -> GaarfBqConfig:
        with open(self.gaarf_config_path, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get("gaarf-bq")
        if not gaarf_section:
            raise ValueError("Invalid config, must have `gaarf-bq` section!")
        params = gaarf_section.get("params")
        return GaarfBqConfig(project=gaarf_section.get("project"),
                             target=gaarf_section.get("target"),
                             params=params)

    def _build_gaarf_config(self) -> GaarfBqConfig:
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(["macro", "sql", "template"]).parse(query_args)
        return GaarfBqConfig(project=main_args.project,
                             target=main_args.dataset,
                             params=params)


class ParamsParser:

    common_params = {"date_iso": datetime.date.today().strftime("%Y%m%d")}

    def __init__(self, identifiers: Sequence[str]):
        self.identifiers = identifiers

    def parse(self,
              params: Sequence[Any]) -> Dict[str, Optional[Dict[str, Any]]]:
        return {
            identifier: self._parse_params(identifier, params)
            for identifier in self.identifiers
        }

    def _parse_params(self, identifier: str,
                      params: Sequence[Any]) -> Dict[Any, Any]:
        parsed_params = {}
        if params:
            raw_params = [param.split("=", maxsplit=1) for param in params]
            for param in raw_params:
                param_pair = self._identify_param_pair(identifier, param)
                if param_pair:
                    parsed_params.update(param_pair)
        return parsed_params

    def _identify_param_pair(self, identifier: str,
                             param: Sequence[str]) -> Optional[Dict[str, Any]]:
        key = param[0]
        if identifier not in key:
            return None
        key = key.replace(f"--{identifier}.", "")
        key = key.replace("-", "_")
        if len(param) == 2:
            return {key: param[1]}
        raise ValueError(f"{identifier} {key} is invalid,"
                         f"--{identifier}.key=value is the correct format")


def convert_date(date_string: str) -> str:
    """Converts specific dates parameters to actual dates."""

    if date_string.find(":YYYY") == -1:
        return date_string
    current_date = datetime.date.today()
    date_object = date_string.split("-")
    base_date = date_object[0]
    if len(date_object) == 2:
        try:
            days_ago = int(date_object[1])
        except ValueError as e:
            raise ValueError(
                "Must provide numeric value for a number lookback period, "
                "i.e. :YYYYMMDD-1") from e
    else:
        days_ago = 0
    if base_date == ":YYYY":
        new_date = datetime.datetime(current_date.year, 1, 1)
        delta = relativedelta(years=days_ago)
    elif base_date == ":YYYYMM":
        new_date = datetime.datetime(current_date.year, current_date.month, 1)
        delta = relativedelta(months=days_ago)
    elif base_date == ":YYYYMMDD":
        new_date = current_date
        delta = relativedelta(days=days_ago)
    return (new_date - delta).strftime("%Y-%m-%d")


class ConfigSaver:

    def __init__(self, path: str):
        self.path = path

    def save(self, gaarf_config: Union[GaarfConfig, GaarfBqConfig]):
        if os.path.exists(self.path):
            with open(self.path, "r", encoding="utf-8") as f:
                config = yaml.safe_load(f)
        else:
            config = {}
        config = self.prepare_config(config, gaarf_config)
        with open(self.path, "w", encoding="utf-8") as f:
            yaml.dump(config,
                      f,
                      default_flow_style=False,
                      sort_keys=False,
                      encoding="utf-8")

    def prepare_config(self, config: Dict[Any, Any],
                       gaarf_config: Union[GaarfConfig, GaarfBqConfig]):
        gaarf = dataclasses.asdict(gaarf_config)
        if isinstance(gaarf_config, GaarfConfig):
            gaarf[gaarf_config.output] = gaarf_config.writer_params
            del gaarf["writer_params"]
            if gaarf_config.writer_params:
                del gaarf["params"][gaarf_config.output]
            gaarf = {k: v for k, v in gaarf.items() if v}
            config.update({"gaarf": gaarf})
        if isinstance(gaarf_config, GaarfBqConfig):
            if (gaarf_bq := config.get("gaarf-bq")):
                if (target := gaarf_bq.get("target")):
                    if target != gaarf_config.target:
                        if isinstance(target, str):
                            target = [target]
                        target = self._handle_target_param(
                            gaarf_config, target)
                    gaarf["target"] = target
            gaarf = {k: v for k, v in gaarf.items() if v}
            config.update({"gaarf-bq": gaarf})
        return config

    @staticmethod
    def _handle_target_param(gaarf_config: GaarfBqConfig,
                             target: List[str]) -> List[str]:
        if isinstance(gaarf_config.target, str):
            target.append(gaarf_config.target)
        elif isinstance(gaarf_config.target, list):
            target.extend(gaarf_config.target)
        return target


def initialize_runtime_parameters(config: Union[GaarfConfig, GaarfBqConfig]):
    for key, param in config.params.items():
        for key_param, value_param in param.items():
            config.params[key][key_param] = convert_date(value_param)
        config.params[key].update(ParamsParser.common_params)
    return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse
from concurrent import futures
from gaarf.io import reader  # type: ignore
from gaarf.bq_executor import BigQueryExecutor
from .utils import GaarfBqConfigBuilder, ConfigSaver, initialize_runtime_parameters


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--project", dest="project")
    parser.add_argument("--target", dest="dataset")
    parser.add_argument("--save-config", dest="save_config", action="store_true")
    parser.add_argument("--no-save-config", dest="save_config", action="store_false")
    parser.add_argument("--config-destination", dest="save_config_dest", default="config.yaml")
    parser.set_defaults(save_config=False)
    args = parser.parse_known_args()
    main_args = args[0]

    config = GaarfBqConfigBuilder(args).build()
    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)

    config = initialize_runtime_parameters(config)

    bq_executor = BigQueryExecutor(config.project)
    bq_executor.create_datasets(config.target)

    reader_client = reader.FileReader()

    with futures.ThreadPoolExecutor() as executor:
        future_to_query = {
            executor.submit(bq_executor.execute, query,
                            reader_client.read(query), config.params): query
            for query in main_args.query
        }
        for future in futures.as_completed(future_to_query):
            query = future_to_query[future]
            try:
                future.result()
                print(f"{query} executed successfully")
            except Exception as e:
                print(f"{query} generated an exception: {e}")


if __name__ == "__main__":
    main()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="google-ads-api-report-fetcher",
    version="0.1.6",
    description=
    "Library for fetching reports from Google Ads API and saving them locally / BigQuery.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/google/ads-api-reports-fetcher",
    author="Google Inc. (gTech gPS CSE team)",
    author_email="no-reply@google.com",
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License"
    ],
    packages=find_packages(),
    install_requires=[
        "google-ads>=16.0.0", "google-cloud-bigquery", "pandas>=1.3.4",
        "pyarrow>=6.0.1", "tabulate", "smart_open[all]", "jinja2", "python-dateutil",
        "sqlalchemy"
    ],
    setup_requires=["pytest-runner"],
    tests_requires=["pytest"],
    entry_points={
        "console_scripts": [
            "gaarf=gaarf.cli.gaarf:main",
            "gaarf-bq=gaarf.cli.bq:main",
        ]
    })



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Sequence
import abc
import proto  # type: ignore


class AbsFormatter(abc.ABC):
    @abc.abstractstaticmethod
    def format(rows: Sequence[Sequence[Any]],
               delimiter: str) -> Sequence[Sequence[Any]]:
        raise NotImplementedError


class ArrayFormatter(AbsFormatter):
    @staticmethod
    def format(rows: Sequence[Sequence[Any]],
               delimiter: str) -> Sequence[Sequence[Any]]:
        formatted_rows = []
        for row in rows:
            formatted_row = []
            for field in row:
                if isinstance(
                        field,
                    (list,
                     proto.marshal.collections.repeated.RepeatedComposite,
                     proto.marshal.collections.repeated.Repeated)):
                    formatted_row.append(
                        delimiter.join([str(element) for element in field]))
                else:
                    formatted_row.append(field)
            formatted_rows.append(formatted_row)
        return formatted_rows


class ResultsFormatter:
    @staticmethod
    def format(results: Sequence[Any]):
        if type(results[0]) in (int, float, str, bool):
            results = [[result] for result in results]
        return results



import pytest
import gaarf.query_editor as query_editor


@pytest.fixture
def query_specification():
    query = """
-- Comment
# Comment
// Comment

SELECT
    customer.id, --customer_id
    campaign.type AS campaign_type, campaign.id:nested AS campaign,
    ad_group.id~1 AS ad_group,
    ad_group_ad.id->asset AS ad,
    campaign.selective_optimization AS selective_optimization,
from ad_group_ad
"""
    return query_editor.QuerySpecification(title="sample_query",
                                           text=query,
                                           args=None)


@pytest.fixture
def sample_query(query_specification):
    return query_specification.generate()


def test_correct_title(sample_query):
    assert sample_query.query_title == "sample_query"


def test_extract_correct_fields(sample_query):
    assert sample_query.fields == [
        "customer.id", "campaign.type_", "campaign.id", "ad_group.id",
        "ad_group_ad.id", "campaign.selective_optimization"
    ]


def test_extract_correct_aliases(sample_query):
    assert sample_query.column_names == [
        "customer_id", "campaign_type", "campaign", "ad_group", "ad",
        "selective_optimization"
    ]


def test_extract_correct_text(sample_query):
    assert sample_query.query_text.lower(
    ) == "select customer.id, campaign.type, campaign.id, ad_group.id, ad_group_ad.id, campaign.selective_optimization from ad_group_ad"


def test_extract_custom_callers(sample_query):
    assert sample_query.customizers == {
        2: {
            "type": "nested_field",
            "value": "nested"
        },
        3: {
            "type": "resource_index",
            "value": 1
        },
        4: {
            "type": "pointer",
            "value": "asset"
        }
    }


def test_format_query(query_specification, sample_query):
    formatted_query = query_specification.extract_query_lines(
        sample_query.query_text)
    assert formatted_query == [
        "customer.id", "campaign.type", "campaign.id", "ad_group.id",
        "ad_group_ad.id", "campaign.selective_optimization"
    ]


def test_extract_correct_resource(sample_query):
    assert sample_query.resource_name == "ad_group_ad"


def test_is_constant_resource(sample_query):
    assert sample_query.is_constant_resource == False
